pub fn pt1(input: &str) -> usize {
    let (ordering_rules, update) = input.split_once("\n\n").unwrap();
    let ordering_rules = ordering_rules
        .lines()
        .map(|line| line.split_once('|').unwrap())
        .map(|(a, b)| (a.parse::<usize>().unwrap(), b.parse::<usize>().unwrap()))
        .collect::<Vec<_>>();

    let updates = update
        .lines()
        .map(|line| {
            line.split(',')
                .map(str::parse::<usize>)
                .map(Result::unwrap)
                .collect::<Vec<_>>()
        })
        .collect::<Vec<_>>();
    updates
        .iter()
        .filter(is_update_sorted(ordering_rules))
        .map(|update| update[update.len() / 2])
        .sum()
}

fn is_update_sorted(ordering_rules: Vec<(usize, usize)>) -> impl FnMut(&&Vec<usize>) -> bool {
    move |update| {
        ordering_rules.iter().all(|(a, b)| {
            if update.contains(a) && update.contains(b) {
                update.iter().position(|u| u == a).unwrap()
                    < update.iter().position(|u| u == b).unwrap()
            } else {
                true
            }
        })
    }
}

pub fn pt2(input: &str) -> usize {
    let (ordering_rules, update) = input.split_once("\n\n").unwrap();
    let ordering_rules = ordering_rules
        .lines()
        .map(|line| line.split_once('|').unwrap())
        .map(|(a, b)| (a.parse::<usize>().unwrap(), b.parse::<usize>().unwrap()))
        .collect::<Vec<_>>();

    let updates = update
        .lines()
        .map(|line| {
            line.split(',')
                .map(str::parse::<usize>)
                .map(Result::unwrap)
                .collect::<Vec<_>>()
        })
        .collect::<Vec<_>>();
    updates
        .iter()
        .filter(|update| {
            ordering_rules.iter().any(|(a, b)| {
                if update.contains(a) && update.contains(b) {
                    update.iter().position(|u| u == a).unwrap()
                        > update.iter().position(|u| u == b).unwrap()
                } else {
                    false
                }
            })
        })
        .map(|update| {
            let mut ordered_update = update.clone();
            let mut sorting_round = 0;
            while !is_update_sorted(ordering_rules.clone())(&&ordered_update) {
                sorting_round += 1;
                tracing::debug!("Round {sorting_round}");
                for (a, b) in &ordering_rules {
                    if ordered_update.contains(a)
                        && ordered_update.contains(b)
                        && ordered_update.iter().position(|u| u == a).unwrap()
                            > ordered_update.iter().position(|u| u == b).unwrap()
                    {
                        let pos_x = ordered_update.iter().position(|u| u == a).unwrap();
                        let pos_y = ordered_update.iter().position(|u| u == b).unwrap();
                        ordered_update.swap(pos_x, pos_y);
                    };
                }
            }
            ordered_update
        })
        .map(|update| update[update.len() / 2])
        .sum()
}

#[cfg(test)]
mod tests {
    use std::fs::read_to_string;

    use super::{pt1, pt2};

    #[test]
    pub fn test_pt1() {
        let input =
            read_to_string("./inputs/day05/example.txt").expect("Missing example.txt for day05");
        assert_eq!(pt1(&input), 143)
    }
    #[test]
    pub fn test_pt2() {
        let input =
            read_to_string("./inputs/day05/example.txt").expect("Missing example.txt for day05");

        assert_eq!(pt2(&input), 123)
    }
}
