use rayon::iter::ParallelBridge;
use rayon::prelude::*;
pub fn pt1(input: &str) -> usize {
    input
        .lines()
        .par_bridge()
        .map(|line| line.split_once(": ").unwrap())
        .map(|(res, args)| {
            (
                res.parse::<usize>().unwrap(),
                args.split(' ')
                    .flat_map(str::parse::<usize>)
                    .collect::<Vec<_>>(),
            )
        })
        .filter(can_calibrate)
        .map(|(val, _)| val)
        .sum()
}
fn can_calibrate((res, args): &(usize, Vec<usize>)) -> bool {
    (0..(2usize << args.len())).any(|mut i| {
        let mut arg_clone = args.iter();
        let mut current_res = *arg_clone.next().unwrap();
        for arg in arg_clone {
            let op = i & 1;
            if op == 1 {
                current_res += arg;
            } else {
                current_res *= arg;
            }
            if current_res > *res {
                return false;
            }
            i >>= 1;
        }
        current_res == *res
    })
}

fn can_calibrate_concat((res, args): &(usize, Vec<usize>)) -> bool {
    (0..(3usize.pow(u32::try_from(args.len()).unwrap()))).any(|mut i| {
        let mut arg_clone = args.iter();
        let mut current_res = *arg_clone.next().unwrap();
        for arg in arg_clone {
            let op = i % 3;
            match op {
                0 => current_res += arg,
                1 => current_res *= arg,
                2 => {
                    current_res *= 10usize.pow(u32::try_from(arg.to_string().len()).unwrap());
                    current_res += arg;
                }
                _ => unreachable!(),
            }
            i /= 3;
            if current_res > *res {
                return false;
            }
        }
        current_res == *res
    })
}
pub fn pt2(input: &str) -> usize {
    input
        .lines()
        .par_bridge()
        .map(|line| line.split_once(": ").unwrap())
        .map(|(res, args)| {
            (
                res.parse::<usize>().unwrap(),
                args.split(' ')
                    .flat_map(str::parse::<usize>)
                    .collect::<Vec<_>>(),
            )
        })
        .filter(can_calibrate_concat)
        .map(|(val, _)| val)
        .sum()
}

#[cfg(test)]
mod tests {
    use super::{pt1, pt2};
    use std::fs::read_to_string;

    #[test]
    pub fn test_pt1() {
        let input =
            read_to_string("./inputs/day07/example.txt").expect("Missing example.txt for day07");
        assert_eq!(pt1(&input), 3749)
    }
    #[test]
    pub fn test_pt2() {
        let input =
            read_to_string("./inputs/day07/example.txt").expect("Missing example.txt for day07");

        assert_eq!(pt2(&input), 11387)
    }
}
