#![allow(
    clippy::cast_sign_loss,
    clippy::cast_possible_wrap,
    clippy::too_many_lines
)]

use std::collections::HashMap;

use tracing::debug;

fn blink_v2(stones: HashMap<usize, usize>) -> HashMap<usize, usize> {
    let mut res: HashMap<usize, usize> = HashMap::new();
    for (stone, count) in stones {
        match stone {
            0 => {
                res.insert(1, *(res.get(&1).unwrap_or(&0)) + count);
            }
            s if s.to_string().len() % 2 == 0 => {
                let s_string = s.to_string();
                let split = s_string.split_at(s_string.len() / 2);
                let left_half = split.0.parse::<usize>().unwrap();
                res.insert(left_half, *(res.get(&left_half).unwrap_or(&0)) + count);
                let right_half = split.1.parse::<usize>().unwrap();
                res.insert(right_half, *(res.get(&right_half).unwrap_or(&0)) + count);
            }
            v => {
                let new_stone = v * 2024;
                res.insert(new_stone, *(res.get(&new_stone).unwrap_or(&0)) + count);
            }
        }
    }
    res
}
pub fn pt1(input: &str) -> usize {
    let mut stones: HashMap<usize, usize> = HashMap::new();
    for stone in input
        .split_whitespace()
        .map(|s| s.parse::<usize>().unwrap())
    {
        stones.insert(stone, stones.get(&stone).unwrap_or(&0) + 1);
    }
    for step in 0..25 {
        debug!("step {step}");
        stones = blink_v2(stones);
    }

    stones.values().sum()
}
pub fn pt2(input: &str) -> usize {
    let mut stones: HashMap<usize, usize> = HashMap::new();
    for stone in input
        .split_whitespace()
        .map(|s| s.parse::<usize>().unwrap())
    {
        stones.insert(stone, stones.get(&stone).unwrap_or(&0) + 1);
    }
    for step in 0..75 {
        debug!("step {step}");
        stones = blink_v2(stones);
    }

    stones.values().sum()
}

#[cfg(test)]
mod tests {
    use tracing_test::traced_test;

    use super::pt1;
    use std::fs::read_to_string;

    #[traced_test]
    #[test]
    pub fn test_pt1() {
        let input =
            read_to_string("./inputs/day11/example.txt").expect("Missing example.txt for day11");
        assert_eq!(pt1(&input), 55312)
    }
}
